'-------------------------------------------------------
' NT Service Example
'
' Original NT Service example in M$asm by:Cynical Pinnacle
' this example beeps every 5 seconds
'
' Ported to SpAsm by
' Eric Asbell (RobotBob) Quanta-IT March 2003
' easbell@quanta-it.com
' Comments are a mix of Pinnacle's and Mine (where noted)
'
' Note: I have added a function called PBStartService
' It will attempt to start a service if it has stopped
' I used it to 'babysit' an unstable email server.
'
' Ported to PowerBasic 7.02 June 2003 by
' Eric Asbell (RobotBob) Quanta-IT March 2003
'-------------------------------------------------------

#COMPILE EXE
#DIM ALL
#INCLUDE "Win32API.inc"

'-------------------------------------------------------
'The name of the service
'-------------------------------------------------------
GLOBAL SERVICE_NAME AS STRING
GLOBAL sTable AS SERVICE_TABLE_ENTRY

'-------------------------------------------------------
' Flags holding current state of service
'-------------------------------------------------------
GLOBAL fPaused AS BYTE
GLOBAL fRunning AS BYTE

'-------------------------------------------------------
' Thread for the actual work
'-------------------------------------------------------
GLOBAL hThread AS DWORD

'-------------------------------------------------------
' Event used to hold ServiceMain from completing
'-------------------------------------------------------
GLOBAL evTerminate AS DWORD

'-------------------------------------------------------
' For QuantaStartService
' only called by that
' function feel free to remove it.
'-------------------------------------------------------
GLOBAL schSCManager AS DWORD
GLOBAL schService AS DWORD
GLOBAL ssStatus AS SERVICE_STATUS

'-----------------------
' For the example
'-----------------------
GLOBAL sStatus AS SERVICE_STATUS

'-------------------------------------------------------
'Handle used to communicate status info with
'the SCM. Created by RegisterServiceCtrlHandler
'-------------------------------------------------------
GLOBAL hStatus AS DWORD

'-------------------------------------------------------
' Program Entry Point
'-------------------------------------------------------
FUNCTION PBMAIN
    fPaused = %FALSE
    fRunning = %FALSE
    ' This is the 'true' name of your service.
    ' Remember this! You need to tell the SCM
    ' this when you install it.
    SERVICE_NAME = "MyService"

    LOCAL rtrval AS DWORD
    LOCAL srvptr AS DWORD

   '----------------------
   'Register with the SCM
   '----------------------
   sTable.lpServiceProc = CODEPTR(ServiceMain)
   srvptr = STRPTR(SERVICE_NAME)
   sTable.lpServiceName = srvptr

   rtrval = StartServiceCtrlDispatcher(sTable)

   IF rtrval = 0 THEN
      CALL ErrorHandler(0)
   END IF

   CALL ExitProcess(rtrval)

END FUNCTION

'-------------------------------------------------------
' This is the main working thread for our service
'-------------------------------------------------------
FUNCTION MainThread(BYVAL var AS LONG) AS LONG
    ' START HERE!
    ' All this does is beep every 5 seconds
    ' Remove this annoying :) example and add your code here
    DO
        SLEEP 5000
        BEEP
    LOOP

    FUNCTION = 0

END FUNCTION

'-------------------------------------------------------
'Initializes the service by starting its thread
'-------------------------------------------------------
FUNCTION Init AS DWORD
   DIM id AS DWORD

   hThread = CreateThread(BYVAL 0, 0, BYVAL CODEPTR(MainThread), 0, 0, id )

   IF hThread <> 0 THEN
      fRunning  = 1
      FUNCTION = 1
   END IF
END FUNCTION

'-------------------------------------------------------
'Resumes a paused primary thread
'-------------------------------------------------------
' Originally this was Resume, but it conflicts with a PB
' keyword. Do whatever you need to RESUME your service
' here
FUNCTION ResumeService AS DWORD
   fPaused = %FALSE

   IF ResumeThread(hThread) = -1 THEN
       ' It failed to resume, handle it if you want.
   END IF
END FUNCTION

'-------------------------------------------------------
' Pauses the primary thread
'-------------------------------------------------------
' Originally this was Pause, but it conflicts with a PB
' keyword. Do whatever you need to SUSPEND your service
' here
FUNCTION PauseService AS DWORD
   fPaused = %TRUE

   IF SuspendThread(hThread) = -1 THEN
    ' It failed to pause, handle it if you want.
   END IF
END FUNCTION

'-------------------------------------------------------
' Stops the service by allowing ServiceMain to
' complete
'-------------------------------------------------------
FUNCTION StopService AS DWORD
   fRunning  = %FALSE
   'This will release ServiceMain
   CALL SetEvent(evTerminate)
END FUNCTION

'-------------------------------------------------------
' Error Handler - Does nothing but bails.
'-------------------------------------------------------
FUNCTION ErrorHandler(srverr AS DWORD) AS DWORD
   ' A more robust error handler could (and should) be made.
   CALL ExitProcess(srverr)
END FUNCTION

'-------------------------------------------------------
' This function consolidates the activities of
' updating the service status with SetsStatus.
'-------------------------------------------------------

FUNCTION SendStatus(dwCurrentState AS DWORD, dwWin32ExitCode AS DWORD,_
                    dwServiceSpecificExitCode AS DWORD, _
                    dwCheckPoint AS DWORD, dwWaitHint AS DWORD) AS DWORD

  DIM success AS DWORD

   sStatus.dwServiceType = %SERVICE_WIN32_OWN_PROCESS
   sStatus.dwCurrentState = dwCurrentState
   IF dwCurrentState = %SERVICE_START_PENDING THEN
      sStatus.dwControlsAccepted  = 0
   ELSE
     sStatus.dwControlsAccepted = %SERVICE_ACCEPT_STOP OR _
     %SERVICE_ACCEPT_PAUSE_CONTINUE OR %SERVICE_ACCEPT_SHUTDOWN
   END IF

   IF dwServiceSpecificExitCode = 0 THEN
      sStatus.dwWin32ExitCode = dwWin32ExitCode
   ELSE
      sStatus.dwWin32ExitCode = %ERROR_SERVICE_SPECIFIC_ERROR
   END IF

   sStatus.dwServiceSpecificExitCode  = dwServiceSpecificExitCode
   sStatus.dwCheckPoint = dwCheckPoint
   sStatus.dwWaitHint = dwWaitHint

   '----------------------------------
   'Pass the status record to the SCM
   '----------------------------------
   CALL SetServiceStatus(hStatus,sStatus)

   '-------------
   ' Return True
   '-------------
   FUNCTION = 1

END FUNCTION

'-------------------------------------------------------
' Dispatches events received from the service
' control manager
'-------------------------------------------------------

SUB CtrlHandler(BYVAL controlCode AS DWORD)

  DIM currentState AS DWORD
  DIM success AS DWORD

  currentState = 0

   IF controlCode = %SERVICE_CONTROL_STOP THEN
      currentState  = %SERVICE_STOP_PENDING

      '-----------------------------
      'Tell the SCM what's happening
      '-----------------------------
      CALL SendStatus(%SERVICE_STOP_PENDING, %NO_ERROR, 0, 1, 5000)

      '--------------------------------
      'Not much to do if not successful
      'Stop the service
      '--------------------------------
      CALL StopService
      EXIT SUB

   '------------------
   'Pause the service
   '------------------
   ELSEIF controlCode = %SERVICE_CONTROL_PAUSE THEN
      IF fRunning <> 0 OR fPaused = 0 THEN
         '-----------------------------
         ' Tell the SCM what's happening
         '-----------------------------
         CALL SendStatus(%SERVICE_PAUSE_PENDING, %NO_ERROR, 0, 1, 1000)
         CALL PauseService
         currentState  = %SERVICE_PAUSED
         GOTO SCHandler
      END IF

   '---------------------
   ' Resume from a pause
   '---------------------
   ELSEIF controlCode = %SERVICE_CONTROL_CONTINUE THEN
       IF fRunning <> 0 OR  fPaused = 1 THEN

         '-------------------------------
         ' Tell the SCM what's happening
         '-------------------------------
          CALL SendStatus(%SERVICE_CONTINUE_PENDING, %NO_ERROR, 0, 1, 1000)
          CALL ResumeService
          currentState  = %SERVICE_RUNNING
          GOTO SCHandler
       END IF

   '-----------------------
   ' Update current status
   '-----------------------
  ELSEIF controlCode = %SERVICE_CONTROL_INTERROGATE THEN
     '-------------------------------------------
     ' it will fall to bottom and send status
     ' Could do cleanup here but it must be
     ' very quick.
     '-------------------------------------------

  ELSEIF controlCode = %SERVICE_CONTROL_SHUTDOWN THEN
     '------------------------------------------
     ' The service is notified when system
     ' shutdown occurs.
     ' Do nothing on shutdown, in this example.
     '------------------------------------------

  END IF

 SCHandler:
     CALL SendStatus(currentState, %NO_ERROR, 0, 0, 0)
END SUB

'-------------------------------------------------------
' Handle an error from ServiceMain by cleaning up
' and telling SCM that the service didn't start.
'-------------------------------------------------------

SUB Terminate(eror AS DWORD)

    DIM result AS DWORD
   '-------------------------------------------
   ' if evTerminate has been created, close it.
   '-------------------------------------------
   IF evTerminate <> 0 THEN
      CALL CloseHandle(evTerminate)
   END IF

    '------------------------------------------
    ' Send a message to the scm to tell about
    ' stopage
    '------------------------------------------
    IF hStatus <> 0 THEN
       CALL SendStatus(%SERVICE_STOPPED,eror, 0, 0, 0)
    END IF

    '--------------------------------------
    ' If the thread has started kill it off
    '--------------------------------------
    IF hThread <> 0 THEN
       CALL CloseHandle(hThread)
    END IF
   '----------------------------
   ' Do not need to close hStatus
   '----------------------------

END SUB

'-------------------------------------------------------
' ServiceMain is called when the SCM wants to
' start the service. When it returns, the service
' has stopped. It therefore waits on an event
' just before the end of the function, and
' that event gets set when it is time to stop.
' It also returns on any error because the
' service cannot start if there is an error.
'-------------------------------------------------------

FUNCTION ServiceMain(argc AS DWORD,argv AS STRING PTR) AS DWORD

   LOCAL success AS DWORD
   LOCAL temp AS DWORD
   LOCAL errnum AS DWORD

   '-----------------------------------------
   ' immediately call Registration function
   '-----------------------------------------
   hStatus = RegisterServiceCtrlHandler(BYVAL STRPTR(SERVICE_NAME),CODEPTR(CtrlHandler))

   IF hStatus = 0 THEN
      errnum = GetLastError
      CALL Terminate(errnum)
      EXIT FUNCTION
   END IF

   '----------------------
   ' Notify SCM of progress
   '----------------------
   CALL SendStatus(%SERVICE_START_PENDING, %NO_ERROR, 0, 1, 5000)

   '----------------------------
   ' create the termination event
   '----------------------------
   evTerminate = CreateEvent(BYVAL %NULL, %TRUE, %FALSE,BYVAL %NULL)

   IF evTerminate = 0 THEN
      errnum = GetLastError
      CALL Terminate(errnum)
      EXIT FUNCTION
   END IF

   '----------------------
   ' Notify SCM of progress
   '----------------------
   CALL SendStatus(%SERVICE_START_PENDING, %NO_ERROR, 0, 2, 1000 )

   '----------------------
   ' Notify SCM of progress
   '----------------------
   CALL SendStatus(%SERVICE_START_PENDING ,%NO_ERROR, 0, 3, 5000)

   '------------------------
   ' Start the service itself
   '------------------------
   success = Init

   IF success = 0 THEN
      errnum = GetLastError
      CALL Terminate(errnum)
      EXIT FUNCTION
   END IF

   '----------------------
   ' Notify SCM of progress
   '----------------------
   CALL SendStatus(%SERVICE_RUNNING, %NO_ERROR, 0, 0, 0)

   '----------------------------------------
   ' Wait for stop signal, and then terminate
   '----------------------------------------
   CALL WaitForSingleObject(evTerminate, %INFINITE)
   CALL Terminate(0)

END FUNCTION

'-------------------------------------------------------
'               PBStartService
'-------------------------------------------------------

FUNCTION PBStartService(program_name AS STRING) AS DWORD

    LOCAL dwStatus AS DWORD
    LOCAL retval AS DWORD

    schSCManager = OpenSCManager($NUL, $NUL, %SC_MANAGER_ALL_ACCESS)

    IF schSCManager = 0 THEN
        EXIT FUNCTION
    END IF

    schService = OpenService(schSCManager, BYVAL STRPTR(program_name), %SERVICE_ALL_ACCESS)

    IF schService = 0 THEN
       EXIT FUNCTION
    END IF

    retval = QueryServiceStatus(schService, ssStatus)
    IF retval = 0 THEN
        EXIT FUNCTION
    END IF

    IF ssStatus.dwCurrentState = %SERVICE_STOPPED THEN
            retval = StartService(schService, 0, 0)
            IF retval = 0 THEN
               EXIT FUNCTION
            END IF
    END IF

    CALL CloseServiceHandle(schService)
    CALL CloseServiceHandle(schSCManager)

END FUNCTION
